package org.infinispan.loaders.database;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.HashSet;
import java.util.Set;

import org.infinispan.Cache;
import org.infinispan.container.entries.InternalCacheEntry;
import org.infinispan.container.entries.InternalEntryFactory;
import org.infinispan.loaders.AbstractCacheLoader;
import org.infinispan.loaders.CacheLoaderConfig;
import org.infinispan.loaders.CacheLoaderException;
import org.infinispan.loaders.CacheLoaderMetadata;
import org.infinispan.loaders.jdbc.connectionfactory.ConnectionFactory;
import org.infinispan.marshall.StreamingMarshaller;
import org.infinispan.marshall.jboss.GenericJBossMarshaller;
import org.infinispan.server.core.CacheValue;
import org.infinispan.util.ByteArrayKey;
import org.infinispan.util.logging.Log;
import org.infinispan.util.logging.LogFactory;


@CacheLoaderMetadata(configurationClass = DataBaseCacheLoaderConfig.class)
public class DataBaseCacheLoader extends AbstractCacheLoader {
    
    private static final Log log = LogFactory.getLog(DataBaseCacheLoader.class);

    private ConnectionFactory connectionFactory;
    private DataBaseCacheLoaderConfig config;
    private TableEntryMapper tableEntryMapper;

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.infinispan.loaders.AbstractCacheLoader#init(org.infinispan.loaders
     * .CacheLoaderConfig, org.infinispan.Cache,
     * org.infinispan.marshall.StreamingMarshaller)
     */
    @Override
    public void init(CacheLoaderConfig config, Cache<?, ?> cache,
	    StreamingMarshaller m) throws CacheLoaderException {
	super.init(config, cache, m);

	this.config = (DataBaseCacheLoaderConfig) config;
	this.connectionFactory = ConnectionFactory.getConnectionFactory(
		getCurrentConfig().getConnectionFactoryConfig()
			.getConnectionFactoryClass(), this.getCurrentConfig()
			.getClassLoader());
	this.marshaller = m;

	try {
	    this.tableEntryMapper = (TableEntryMapper) Class.forName(this.config.getTableEntryMapperClass()).newInstance();
	} catch (Exception e) {
	    throw new CacheLoaderException("TableEntryMapper could not be instantiated. Class: " + this.config.getTableEntryMapperClass(), e);
	}

	log.debug("DataBaseCacheLoader initialized");
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.infinispan.loaders.CacheLoader#start()
     */
    public void start() throws CacheLoaderException {
	log.debug("Starting DataBaseCacheLoader");
	getConnectionFactory().start(
		this.getCurrentConfig().getConnectionFactoryConfig(),
		this.getCurrentConfig().getClassLoader());
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.infinispan.loaders.CacheLoader#stop()
     */
    public void stop() throws CacheLoaderException {
	log.debug("Stoping DataBaseCacheLoader");
	getConnectionFactory().stop();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.infinispan.loaders.CacheLoader#load(java.lang.Object)
     */
    public InternalCacheEntry load(Object key) throws CacheLoaderException {
	if (!(key instanceof ByteArrayKey)) {
	    return null;
	}

	Object convertedKey = null;

	try {
	    convertedKey = convertKey((ByteArrayKey) key);
	} catch (Exception e) {
	    throw new CacheLoaderException("Key could not be converted to its representation.", e);
	}

	Object databaseEntry = null;

	try {
	    databaseEntry = this.loadFromDatabase(convertedKey);

	    if (databaseEntry != null) {
		byte[] objectToByteBuffer = this.marshaller.objectToByteBuffer(databaseEntry);

		return InternalEntryFactory.create((ByteArrayKey) key, new CacheValue(objectToByteBuffer, 1));
	    }
	} catch (Exception e) {
	    throw new CacheLoaderException("Error while marshalling entry instance. Instance: " + databaseEntry, e);
	}

	return null;
    }

    public Set<InternalCacheEntry> loadAll() throws CacheLoaderException {
	return new HashSet<InternalCacheEntry>();
    }

    public Set<InternalCacheEntry> load(int numEntries)
	    throws CacheLoaderException {
	return new HashSet<InternalCacheEntry>();
    }

    public Set<Object> loadAllKeys(Set<Object> keysToExclude)
	    throws CacheLoaderException {
	return new HashSet<Object>();
    }

    public Class<? extends CacheLoaderConfig> getConfigurationClass() {
	return DataBaseCacheLoaderConfig.class;
    }

    private Object loadFromDatabase(Object convertedKey)
	    throws CacheLoaderException {
	log.debug("Loading " + convertedKey + " from database.");

	Connection connection = null;
	PreparedStatement pst = null;
	ResultSet rs = null;

	try {
	    connection = getConnectionFactory().getConnection();

	    pst = connection.prepareStatement(this.getCurrentConfig()
		    .getQuery());

	    pst.setObject(1, convertedKey);

	    log.debug("Executing query " + this.getCurrentConfig().getQuery());

	    rs = pst.executeQuery();

	    if (rs.next()) {
		log.debug("Entry was found. Converting it using mapper: "
			+ this.tableEntryMapper);

		return this.tableEntryMapper.transform(rs);
	    }
	} catch (Exception e) {
	    throw new CacheLoaderException("Error while loading key "
		    + convertedKey + " from database.", e);
	} finally {
	    try {
		rs.close();
		pst.close();
		getConnectionFactory().releaseConnection(connection);
	    } catch (Exception e) {
		log.error("Error while loading key " + convertedKey
			+ " from database.", e);
	    }
	}

	log.debug("Entry does not exists in database.Key: " + convertedKey);

	return null;
    }

    private DataBaseCacheLoaderConfig getCurrentConfig() {
	return this.config;
    }

    private Object convertKey(ByteArrayKey key) throws IOException,
	    ClassNotFoundException {
	try {
	    return this.marshaller.objectFromByteBuffer(key.getData());
	} catch (Exception e) {
	    return new GenericJBossMarshaller().objectFromByteBuffer(key
		    .getData());
	}
    }

    private ConnectionFactory getConnectionFactory() {
	return this.connectionFactory;
    }

}
