package test.org.infinispan.loaders.database;

import static org.testng.Assert.assertNotNull;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

import org.infinispan.CacheImpl;
import org.infinispan.container.entries.InternalCacheEntry;
import org.infinispan.loaders.CacheLoaderException;
import org.infinispan.loaders.database.DataBaseCacheLoader;
import org.infinispan.loaders.database.DataBaseCacheLoaderConfig;
import org.infinispan.loaders.jdbc.connectionfactory.ConnectionFactoryConfig;
import org.infinispan.loaders.jdbc.connectionfactory.PooledConnectionFactory;
import org.infinispan.marshall.VersionAwareMarshaller;
import org.infinispan.test.fwk.UnitTestDatabaseManager;
import org.infinispan.util.ByteArrayKey;
import org.testng.annotations.AfterTest;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.Test;

public class TestDataBaseLoader {

    private static final String DEFAULT_JDBC_URL = "jdbc:h2:file:/tmp/testdb";
    private static final String SQL_INSERT_TABLE_ENTRY = "INSERT INTO tb_cache_entry (field1, field2, field3) VALUES (?,?,?)";
    private static final String SQL_CREATE_MOCK_TABLE = "CREATE TABLE tb_cache_entry (id INTEGER IDENTITY, field1 VARCHAR(255), field2 VARCHAR(255), field3 VARCHAR(255))";
    private static final String SQL_DROP_MOCK_TABLE = "DROP TABLE tb_cache_entry";

    private DataBaseCacheLoader dataBaseCacheLoader;
    private PooledConnectionFactory pooledConnectionFactory;
    
    private VersionAwareMarshaller marshaller = new VersionAwareMarshaller();
    
    @BeforeTest
    public void onSetup() throws CacheLoaderException, SQLException {
	createDatabaseCacheLoader();

	ConnectionFactoryConfig uniqueConnectionFactoryConfig = UnitTestDatabaseManager.getUniqueConnectionFactoryConfig();

	uniqueConnectionFactoryConfig.setConnectionUrl(DEFAULT_JDBC_URL);
	
	this.pooledConnectionFactory = new PooledConnectionFactory();
	
	this.pooledConnectionFactory.start(uniqueConnectionFactoryConfig, Thread.currentThread().getContextClassLoader());

	Connection connection = this.pooledConnectionFactory.getConnection();

	loadEntriesOnDatabase(connection);

	this.pooledConnectionFactory.releaseConnection(connection);
    }
    
    @AfterTest
    public void onFinish() throws CacheLoaderException {
	this.dataBaseCacheLoader.stop();
	this.pooledConnectionFactory.stop();
    }

    @Test
    public void testLoad() throws CacheLoaderException, IOException,
	    InterruptedException, SQLException {
	InternalCacheEntry loadedEntry = this.dataBaseCacheLoader.load(new ByteArrayKey(marshaller.objectToByteBuffer(1l)));

	assertNotNull(loadedEntry);
    }

    private void loadEntriesOnDatabase(Connection connection) throws CacheLoaderException, SQLException {
	PreparedStatement dropTableStatement = connection.prepareStatement(SQL_DROP_MOCK_TABLE);

	dropTableStatement.execute();
	dropTableStatement.close();

	PreparedStatement createTableStatement = connection.prepareStatement(SQL_CREATE_MOCK_TABLE);

	createTableStatement.execute();
	createTableStatement.close();

	PreparedStatement insertStatement = connection.prepareStatement(SQL_INSERT_TABLE_ENTRY);

	insertStatement.setString(1, "field1");
	insertStatement.setString(2, "field2");
	insertStatement.setString(3, "field3");

	insertStatement.execute();
	insertStatement.close();
    }

    private void createDatabaseCacheLoader() throws CacheLoaderException {
	this.dataBaseCacheLoader = new DataBaseCacheLoader();

	this.dataBaseCacheLoader.init(createDataBaseCacheLoaderConfig(), new CacheImpl<Object, Object>("default"), this.marshaller);

	this.dataBaseCacheLoader.start();
    }
    
    private DataBaseCacheLoaderConfig createDataBaseCacheLoaderConfig() {
	ConnectionFactoryConfig uniqueConnectionFactoryConfig = UnitTestDatabaseManager.getUniqueConnectionFactoryConfig();

	uniqueConnectionFactoryConfig.setConnectionUrl(DEFAULT_JDBC_URL);

	DataBaseCacheLoaderConfig dataBaseCacheLoaderConfig = new DataBaseCacheLoaderConfig(uniqueConnectionFactoryConfig);

	dataBaseCacheLoaderConfig.setConnectionFactoryClass("org.infinispan.loaders.jdbc.connectionfactory.PooledConnectionFactory");
	dataBaseCacheLoaderConfig.setTableEntryMapperClass("test.org.infinispan.loaders.database.MockTableEntryMapper");
	dataBaseCacheLoaderConfig.setQuery("select id, field1, field2, field3 from tb_cache_entry where id = ?");

	return dataBaseCacheLoaderConfig;
    }
}
